require "algorithm_selector/version"

# algorithm-selector library; 5 methods (all, best, analyze, compare, sort)
module AlgorithmSelector
  extend self

  # Returns the times for each algorithm
  def all(action, data_set, trials=1, target=nil)
    if action == "sort"
      Sorts.new.all(data_set, trials)
    elsif action == "search"
      Searches.new.all(data_set, trials, target)
    end
  end

  # Returns the algorithm with the best time
  def best(action, data_set, trials=1, target=nil)
    if action == "sort"
      Sorts.new.best(data_set, trials)
    elsif action == "search"
      Searches.new.best(data_set, trials, target)
    end
  end

  # Returns time for specified algorithm or data structure
  def analyze(action, data_set, algorithm, trials=1, target=nil)
    if action == "sort"
      Sorts.new.analyze(data_set, algorithm, trials)
    elsif action == "search"
      Searches.new.analyze(data_set, algorithm, trials, target)
    end
  end

  # Returns the times of two specified algorithms for comparison
  def compare(action, data_set, first_algorithm, second_algorithm, trials=1, target=nil)
    if action == "sort"
      Sorts.new.compare(data_set, first_algorithm, second_algorithm, trials)
    elsif action == "search"
      Searches.new.compare(data_set, first_algorithm, second_algorithm, trials, target)
    end
  end

  # Returns sorted data_set of specified sorting algorithm
  def sort(data_set, algorithm)
    method = Sorts.new.method(algorithm)
    method.call(data_set)
  end

end

# Sorts class
# 4 methods corresponding to algorithm_selector module
# 5 sorting algorithms (selection, insertion, bubble, merge, quick)
class Sorts

  # Show all results from sorting algorithms
  def all(data_set, trials)
    {
      sorted: merge_sort(data_set),
      results: {
        bubble_sort: display_time(Proc.new {bubble_sort(data_set)}, trials),
        insertion_sort: display_time(Proc.new {insertion_sort(data_set)}, trials),
        selection_sort: display_time(Proc.new {selection_sort(data_set)}, trials),
        merge_sort: display_time(Proc.new {merge_sort(data_set)}, trials),
        quick_sort: display_time(Proc.new {quick_sort(data_set)}, trials),
        heap_sort: display_time(Proc.new {heap_sort(data_set)}, trials)
      }
    }
  end

  # Show the best sorting algorithm
  def best(data_set, trials)
    hash = all(data_set, trials)
    best_time = 0
    best_algorithm = {}
    hash[:results].each do |key, val|
      if val < best_time || best_time == 0
        best_time = val
        best_algorithm = Hash[key, val]
      end
    end
    {
      best_algorithm.keys[0] => best_algorithm[best_algorithm.keys[0]]
    }
  end

  # Show specific sorting algorithm by name
  def analyze(data_set, algorithm, trials)
    hash = all(data_set, trials)
    alg = {}
    hash[:results].each do |key, val|
      if algorithm.to_sym == key
        alg = Hash[key, val]
      end
    end
    {
      sorted: hash[:sorted],
      alg.keys[0] => alg[alg.keys[0]]
    }
  end

  # Show comparison of two algorithms
  def compare(data_set, first_algorithm, second_algorithm, trials)
    hash = all(data_set, trials)
    first = {}
    second = {}
    hash[:results].each do |key, val|
      if first_algorithm.to_sym == key
        first = Hash[key, val]
      elsif second_algorithm.to_sym == key
        second = Hash[key, val]
      end
    end
    {
      first.keys[0] => first[first.keys[0]],
      second.keys[0] => second[second.keys[0]]
    }
  end

  # Bubble sort
  # Worst-Case Space Complexity: O(1)
  # Averge-Case Time Complexity: O(n^2)
  # Worst-Case Time Complexity: O(n^2)
  def bubble_sort(data_set)
    return data_set if data_set.length < 2
    switched = true
    until switched == false
      switched = false
      (1...data_set.length).each do |i|
        if data_set[i-1] > data_set[i]
          temp = data_set[i]
          data_set[i] = data_set[i-1]
          data_set[i-1] = temp
          switched = true
        end
      end
    end
    return data_set
  end

  # Insertion sort
  # Worst-Case Space Complexity: O(1)
  # Averge-Case Time Complexity: O(n^2)
  # Worst-Case Time Complexity: O(n^2)
  def insertion_sort(data_set)
    return data_set if data_set.length < 2
    (1...data_set.length).to_a do |i|
      value = data_set[i]
      j = i - 1
      while (j >= 0 && data_set[j] > value) do
        data_set[j+1] = data_set[j]
        j = j - 1
      end
      data_set[j+1] = value
    end
    data_set
  end

  # Selection sort
  # Worst-Case Space Complexity: O(1)
  # Averge-Case Time Complexity: O(n^2)
  # Worst-Case Time Complexity: O(n^2)
  def selection_sort(data_set)
    (0...data_set.length-1).to_a.each do |i|
      min = i
      (i+1...data_set.length).to_a.each do |j|
        if (data_set[j] < data_set[min])
          min = j
        end
      end
      if min != i
        data_set[min], data_set[i] = data_set[i], data_set[min]
      end
    end
    return data_set
  end

  # Merge sort
  # Worst-Case Space Complexity: O(n)
  # Averge-Case Time Complexity: O(n log(n))
  # Worst-Case Time Complexity: O(n log(n))
  def merge_sort(data_set)
    return data_set if data_set.length < 2
    middle = data_set.length/2
    left = data_set[0...middle]
    right = data_set[middle...data_set.length]
    merge(merge_sort(left), merge_sort(right))
  end

  # Quick sort
  # Worst-Case Space Complexity: O(log(n))
  # Averge-Case Time Complexity: O(n log(n))
  # Worst-Case Time Complexity: O(n^2)
  def quick_sort(data_set)
    return data_set if data_set.count < 2
    pivot = data_set[0]
    left = data_set.select {|el| el < pivot}
    right = data_set.select {|el| el > pivot}
    quick_sort(left) + [pivot] + quick_sort(right)
  end

  # Quick sort
  # Worst-Case Space Complexity: O(1)
  # Averge-Case Time Complexity: O(n log(n))
  # Worst-Case Time Complexity: O(n log(n))
  def heap_sort(data_set)
    (2..data_set.length).to_a.each do |idx|
      BinaryMinHeap.heapify_up(data_set, idx - 1, idx)
    end
    (2..data_set.length).to_a.reverse.each do |idx|
      data_set[idx - 1], data_set[0] = data_set[0], data_set[idx - 1]
      BinaryMinHeap.heapify_down(data_set, 0, idx - 1)
    end
    data_set.reverse!
  end
end

# Searches class
# 4 methods corresponding to algorithm_selector module
class Searches

  # Show all data structures for searching
  def all(data_set, trials, target)
    stack = Stack.new(data_set.length, data_set)
    queue = Queue.new(data_set.length, data_set)
    linked_list = LinkedList.new(data_set)
    binary_tree = BinaryTree.new(data_set)
    {
      found: stack.search(target),
      results:
        {
          stack: display_time(Proc.new {stack.search(target)}, trials),
          queue: display_time(Proc.new {queue.search(target)}, trials),
          linked_list: display_time(Proc.new {linked_list.search(target)}, trials),
          binary_tree: display_time(Proc.new {binary_tree.search(target)}, trials)
        }
    }
  end

  # Show best data structure for searching
  def best(data_set, trials, target)
    hash = all(data_set, trials, target)
    best_time = 0
    best_algorithm = {}
    hash[:results].each do |key, val|
      if val < best_time || best_time == 0
        best_time = val
        best_algorithm = Hash[key, val]
      end
    end
    {
      best_algorithm.keys[0] => best_algorithm[best_algorithm.keys[0]]
    }
  end

  # Show analysis of a data structure for searching
  def analyze(data_set, algorithm, trials, target)
    hash = all(data_set, trials, target)
    alg = {}
    hash[:results].each do |key, val|
      if algorithm.to_sym == key
        alg = Hash[key, val]
      end
    end
    {
      found: hash[:found],
      alg.keys[0] => alg[alg.keys[0]]
    }
  end

  # Show comparison of two data structures for searching
  def compare(data_set, first_algorithm, second_algorithm, trials, target)
    hash = all(data_set, trials, target)
    first = {}
    second = {}
    hash[:results].each do |key, val|
      if first_algorithm.to_sym == key
        first = Hash[key, val]
      elsif second_algorithm.to_sym == key
        second = Hash[key, val]
      end
    end
    {
      first.keys[0] => first[first.keys[0]],
      second.keys[0] => second[second.keys[0]]
    }
  end
end

# Stack data structure
# Worst-Case Space Complexity: O(n)
class Stack
  def initialize(size, data_set=[])
    @size = size
    @store = Array.new(@size)
    @top = -1
    data_set.each { |el| push(el) }
  end

  # Search, Average-Case Time Complexity: O(n)
  # Search, Worst-Case Time Complexity: O(n)
  def search(target)
    found = false
    new_stack = Stack.new(@size)
    until empty? do
      new_stack.push(pop)
      if new_stack.look == target
        found = true
        break
      end
    end
    push(new_stack.pop) until new_stack.empty?
    found
  end

  def pop
    if empty?
      nil
    else
      popped = @store[@top]
      @store[@top] = nil
      @top = @top.pred
      popped
    end
  end

  # Insertion, Worst-Case Time Complexity: O(1)
  def push(el)
    if full? or el.nil?
      nil
    else
      @top = @top.succ
      @store[@top] = el
      self
    end
  end

  def size
    @size
  end

  def look
    @store[@top]
  end

  def full?
    @top == (@size - 1)
  end

  def empty?
    @top == -1
  end
end

# Queue data structure
# Worst-Case Space Complexity: O(n)
class Queue
  def initialize(size, data_set=[])
    @size = size
    @store = Array.new(@size)
    @head, @tail = -1, 0

    # initialize Queue with data_set if given
    data_set.each { |el| enqueue(el) }
  end

  # Search, Worst-Case Time Complexity: O(n)
  def search(target)
    found = false
    @size.times do
      found = true if look == target
      enqueue(dequeue)
    end
    found
  end

  def dequeue
    if empty?
      nil
    else
      @tail = @tail.succ
      dequeued = @store[@head]
      @store.unshift(nil)
      @store.pop
      dequeued
    end
  end

  # Insertion, Worst-Case Time Complexity: O(1)
  def enqueue(el)
    if full? or el.nil?
      nil
    else
      @tail = @tail.pred
      @store[@tail] = el
      self
    end
  end

  def size
    @size
  end

  def look
    @store[@head]
  end

  private

  def empty?
    @head == -1 and @tail == 0
  end

  def full?
    @tail.abs == (@size)
  end
end

# Singly-Linked List data structure
# Worst-Case Space Complexity: O(n)
class LinkedList
  include Enumerable

  def initialize(data_set=[])
    @head = LinkNode.new
    @tail = LinkNode.new
    @head.next = @tail
    @tail.prev = @head
    data_set.each { |i| insert(i) }
  end

  # Search, Worst-Case Time Complexity: O(n)
  def search(target)
    each { |current_link| return current_link.val if current_link.val == target }
  end

  def [](i)
    each_with_index { |link, j| return link if i == j }
    nil
  end

  def first
    empty? ? nil : @head.next
  end

  def last
    empty? ? nil : @tail.prev
  end

  def empty?
    @head.next == @tail
  end

  def get(val)
    each { |link| return link.val if link.val == val }
    nil
  end

  def include?(val)
    any? { |link| link.val == val }
  end

  # Insertion, Worst-Case Time Complexity: O(1)
  def insert(val)
    each { |link| return link.val = val if link.val == val }
    new_link = LinkNode.new(val)
    @tail.prev.next = new_link
    new_link.prev = @tail.prev
    new_link.next = @tail
    @tail.prev = new_link
    new_link
  end

  # Deletion, Worst-Case Time Complexity: O(1)
  def remove(val)
    each do |link|
      if link.val == val
        link.prev.next = link.next
        link.next.prev = link.prev
        link.next, link.prev = nil, nil
        return link.val
      end
    end
    nil
  end

  def each
    current_link = @head.next
    until current_link == @tail
      yield current_link
      current_link = current_link.next
    end
  end
end

# Binary Search Tree data structure
# Worst-Case Space Complexity: O(n)
class BinaryTree
  attr_accessor :root

  def initialize(data_set=[], root_val=nil)
    @root = BSTNode.new(root_val)
    data_set.each { |i| insert(i) }
  end

  # Search, Worst-Case Time Complexity: O(n)
  def search(target)
    return nil if @root.nil?
    return search_in_tree(@root, target)
  end

  def search_in_tree(tree_node, target)
    return nil if tree_node.nil?
    return tree_node.val if target == tree_node.val
    if target < tree_node.val
      search_in_tree(tree_node.left, target)
    else
      search_in_tree(tree_node.right, target)
    end
  end

  # Insertion, Worst-Case Time Complexity: O(n)
  def insert(val)
    new_node = BSTNode.new(val)
    @root = new_node and return unless @root.val
    insert_to_tree(@root, new_node)
  end

  def insert_to_tree(tree_node, new_node)
    tree_node = new_node and return if tree_node.nil?
    return tree_node if tree_node.val == new_node.val
    if new_node.val > tree_node.val
      insert_to_tree(tree_node.right, new_node) unless tree_node.right.nil?
      tree_node.right = new_node
    else
      insert_to_tree(tree_node.left, new_node) unless tree_node.left.nil?
      tree_node.left = new_node
    end
  end
end

private

# Helper method to get time required for method to run
# Returns time in seconds
def display_time(prc, trials)
  total = 0
  trials.times do
    start = Time.now
    prc.call
    finish = Time.now
    total = total + (finish - start)
  end
  total/trials
end

# Helper method for merge_sort
def merge(left, right)
  data_set = []
  until left.length == 0 || right.length == 0
    if left.first > right.first
      data_set << right.shift
    else
      data_set << left.shift
    end
  end
  data_set.concat(left).concat(right)
end

# Node helper class for Linked List
class LinkNode
  attr_accessor :val, :next, :prev
  def initialize(val=nil)
    @val = val
    @next = nil
    @prev = nil
  end
end

# Node helper class for Binary Search Tree
class BSTNode
  attr_accessor :val, :left, :right
  def initialize(val=nil, left=nil, right=nil)
    @val = val
    @left = left
    @right = right
  end
end

# MinHeap helper class for Heap Sort
class BinaryMinHeap
  def initialize(&prc)
    @store = []
    @prc = prc || Proc.new { |el1, el2| el1 <=> el2 }
  end

  def count
    @store.length
  end

  def extract
    raise "no element to extract" if count == 0
    val = @store[0]
    @store.pop if count == 1
    if count > 1
      @store[0] = @store.pop
      self.class.heapify_down(@store, 0, &@prc)
    end
    val
  end

  def peek
    raise "no element to peek" if count == 0
    @store[0]
  end

  def push(val)
    @store.push(val)
    self.class.heapify_up(@store, count - 1, &@prc)
  end

  protected
  attr_accessor :prc, :store

  public
  def self.child_indices(len, parent_index)
    [2 * parent_index + 1, 2 * parent_index + 2].select { |idx| idx < len }
  end

  def self.parent_index(child_index)
    raise "root has no parent" if child_index == 0
    (child_index - 1) / 2
  end

  def self.heapify_down(array, parent_idx, len = array.length, &prc)
    prc ||= Proc.new { |el1, el2| el1 <=> el2 }

    left_child, right_child = child_indices(len, parent_idx)

    parent = array[parent_idx]

    children = []
    children.push(array[left_child]) if left_child
    children.push(array[right_child]) if right_child

    if children.all? { |child| prc.call(parent, child) <= 0 }
      return array
    end

    swap_idx = nil
    if children.length == 1
      swap_idx = left_child
    else
      swap_idx =
        prc.call(children[0], children[1]) == -1 ? left_child : right_child
    end

    array[parent_idx], array[swap_idx] = array[swap_idx], parent
    heapify_down(array, swap_idx, len, &prc)
  end

  def self.heapify_up(array, child_idx, len = array.length, &prc)
    prc ||= Proc.new { |el1, el2| el1 <=> el2 }
    return array if child_idx == 0

    parent_idx = parent_index(child_idx)
    child, parent = array[child_idx], array[parent_idx]
    if prc.call(child, parent) >= 0
      return array
    else
      array[child_idx], array[parent_idx] = parent, child
      heapify_up(array, parent_idx, len, &prc)
    end
  end
end
